// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Nathan van Beelen
//
// This file is part of matrix-dart-sdk-sqflite.
//
// matrix-dart-sdk-sqflite is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk-sqflite is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk-sqflite.  If not, see <https://www.gnu.org/licenses/>.

library matrix_sdk_sqflite;

import 'dart:convert';

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:meta/meta.dart';
import 'package:sqflite/sqflite.dart';

class SqfliteStore implements Store {

  final String path;
  SqfliteStore({@required this.path});

  bool _isOpen = false;

  @override
  bool get isOpen => _isOpen;

  LocalUser _user;

  Database _db;

  static const _userInfoTable = 'user_info';
  static const _roomTable = 'room';
  static const _eventTable = 'event';
  static const _userTable = 'user';
  static const _memberTable = 'member';

  static const _idColumn = 'id';

  static const _userIdColumn = 'user_id';
  static const _accessTokenColumn = 'access_token';
  static const _syncTokenColumn = 'sync_token';
  static const _homeserverColumn = 'homeserver';

  static const _roomIdColumn = 'room_id';

  // Room columns
  static const _typeColumn = 'type';
  static const _nameColumn = 'name';
  static const _avatarUrlColumn = 'avatar_url';
  static const _previousBatchColumn = 'previous_batch';

  // Event columns
  static const _contentColumn = 'content';
  static const _senderIdColumn = 'sender_id';
  static const _timeColumn = 'time';

  // Member columns
  static const _displayNameColumn = 'display_name';

  Future<void> _createDatabase(Database db, LocalUser user) async {
    if (user == null) {
      throw ArgumentError(
        'user cannot be null if the database does not exist'
      );
    }

    _user = user;

    final batch = db.batch()
    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_userInfoTable (
        $_idColumn INTEGER PRIMARY KEY,
        $_userIdColumn TEXT,
        $_accessTokenColumn TEXT,
        $_syncTokenColumn TEXT,
        $_homeserverColumn TEXT);
      ''')

    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_roomTable (
          $_idColumn TEXT PRIMARY KEY,
          $_typeColumn TEXT,
          $_nameColumn TEXT NULL,
          $_avatarUrlColumn TEXT NULL,
          $_previousBatchColumn TEXT NOT NULL);
      ''')

    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_eventTable (
          $_idColumn TEXT PRIMARY KEY,
          $_roomIdColumn TEXT,
          $_senderIdColumn TEXT NULL,
          $_timeColumn INT NOT NULL,
          $_contentColumn TEXT NOT NULL,
          FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn));
      ''')

    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_userTable (
          $_idColumn TEXT PRIMARY KEY,
          $_displayNameColumn TEXT NULL);
      ''')

    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_memberTable (
          $_userIdColumn TEXT,
          $_roomIdColumn TEXT,
          FOREIGN KEY($_userIdColumn) REFERENCES $_userTable($_idColumn),
          FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn)
          PRIMARY KEY ($_userIdColumn, $_roomIdColumn));
      ''');

    await batch.commit(noResult: true);
  }

  Future<LocalUser> userFromDb(Database db) async {
    final userId = await getUserId(db: db);
    final hs = await getHomeserver(db: db);
    final accessToken = await getAccessToken(db: db);

    return LocalUser(
        id: userId,
        homeserver: hs,
        accessToken: accessToken,
        store: this
    );
  }

  @override
  Future<bool> open({LocalUser user}) async {
    try {
      _db = await openDatabase(
        path,
        version: 5,
        onUpgrade: (Database db, int oldVersion, int newVersion) async {

          // First get the user from the old database
          final localUser = await userFromDb(db);

          // Drop all tables and delete the sync token
          var batch = db.batch()
            ..execute('UPDATE $_userInfoTable SET $_syncTokenColumn=NULL')
            ..execute('DROP TABLE IF EXISTS $_roomTable')
            ..execute('DROP TABLE IF EXISTS $_eventTable')
            ..execute('DROP TABLE IF EXISTS $_userTable')
            ..execute('DROP TABLE IF EXISTS $_memberTable');

          await batch.commit(noResult: true);

          await _createDatabase(db, localUser);
        },
        onCreate: (Database db, int version) async {
          await _createDatabase(db, user);
        }
      );

      if (_user == null) {
        _user = await userFromDb(_db);
      }
    } on ArgumentError {
      return false;
    }

    _isOpen = true;
    return true;
  }

  @override
  Future<void> close() async {
    _isOpen = false;
  }

  Future<String> _getUserInfo(String column, {Database db}) async {
    if (db == null) {
      db = _db;
    }

    var results = await db.rawQuery(
      '''
      SELECT $column FROM $_userInfoTable WHERE $_idColumn = 1;
      '''
    );

    if (results.isNotEmpty) {
      return results.first[column];
    } else {
      return null;
    }
  }

  Future<void> _setUserInfo(String column, String value, {Database db}) async {
    if (db == null) {
      db = _db;
    }

    final batch = db.batch()
    ..execute(
      '''
      UPDATE $_userInfoTable SET $column = ?;
      ''',
      [value]
    )

    ..execute(
      '''
      INSERT OR IGNORE INTO $_userInfoTable($_idColumn, $column)
        values (?, ?);
      ''',
      [1, value]
    );

    await batch.commit(noResult: true);
  }

  @override
  Future<UserId> getUserId({Database db}) async {
    final value = await _getUserInfo(_userIdColumn, db: db);
    if (value == null) {
      return null;
    }

    return UserId(value);
  }

  @override
  Future<void> setUserId(UserId userId) async {
    await _setUserInfo(_userIdColumn, userId.toString());
  }

  @override
  Future<String> getAccessToken({Database db}) async {
    return _getUserInfo(_accessTokenColumn, db: db);
  }

  @override
  Future<void> setAccessToken(String accessToken) async {
    await _setUserInfo(_accessTokenColumn, accessToken);
  }

  @override
  Future<String> getSyncToken() async {
    return _getUserInfo(_syncTokenColumn);
  }

  @override
  Future<void> setSyncToken(String syncToken) async {
    await _setUserInfo(_syncTokenColumn, syncToken);
  }

  @override
  Future<Homeserver> getHomeserver({Database db}) async {
    final value = await _getUserInfo(_homeserverColumn, db: db);
    if (value == null) {
      return null;
    }

    return Homeserver(Uri.parse(value));
  }

  @override
  Future<void> setHomeserver(Homeserver homeserver) async {
    await _setUserInfo(_homeserverColumn, homeserver.uri.toString());
  }

  @override
  Future<JoinedRoom> getJoinedRoom(RoomId id) async {
    var results = await _db.rawQuery(
      '''
      SELECT $_nameColumn,
             $_avatarUrlColumn,
             $_previousBatchColumn
      FROM $_roomTable WHERE $_idColumn = ?;
      ''',
      [id.toString()]
    );

    if (results.isNotEmpty) {
      final first = results.first;
      final name = first[_nameColumn];
      final avatarUrl = first[_avatarUrlColumn];
      final previousBatch = first[_previousBatchColumn];

      return JoinedRoom(id,
        store: this,
        name: name,
        avatarUrl: avatarUrl,
        previousBatch: previousBatch,
        scope: _user.scope
      );
    } else {
      return null;
    }
  }

  @override
  Future<void> setJoinedRooms(Iterable<JoinedRoom> rooms) async {
    await _db.transaction((txn) {
      for (JoinedRoom room in rooms) {
        var url;
        if (room.avatarUrl != null) {
          url = room.avatarUrl.toString();
        }

        txn.execute(
          '''
          UPDATE $_roomTable SET $_typeColumn = ?,
                                 $_nameColumn = ?,
                                 $_avatarUrlColumn = ?,
                                 $_previousBatchColumn = ?
                             WHERE $_idColumn = ?;
          ''',
          ['join', room.name, url, room.id.toString(), room.previousBatch]
        );

        txn.execute(
          '''
          INSERT OR IGNORE INTO $_roomTable($_idColumn,
                                            $_typeColumn,
                                            $_nameColumn,
                                            $_avatarUrlColumn,
                                            $_previousBatchColumn)
            values (?, ?, ?, ?, ?);
          ''',
          [room.id.toString(), 'join', room.name, url, room.previousBatch]
        );
      }
    });
  }

  @override
  Stream<JoinedRoom> getJoinedRooms() async* {
    var results = await _db.rawQuery(
      '''
      SELECT $_idColumn,
             $_nameColumn,
             $_avatarUrlColumn,
             $_previousBatchColumn
      FROM $_roomTable;
      '''
    );

    if (results.isNotEmpty) {

      for (var result in results) {
        final id = RoomId(result[_idColumn]);
        final name = result[_nameColumn];
        var avatarUrl = result[_avatarUrlColumn];
        final previousBatch = result[_previousBatchColumn];
        if (avatarUrl != null) {
          avatarUrl = Uri.parse(result[_avatarUrlColumn]);
        }
        yield JoinedRoom(id,
          store: this,
          name: name,
          avatarUrl: avatarUrl,
          previousBatch: previousBatch,
          scope: _user.scope
        );
      }
    }
  }

  @override
  Future<Event> getEvent(Room room, EventId id) async {
    var results = await _db.rawQuery(
      '''
      SELECT $_contentColumn FROM $_eventTable WHERE $_idColumn = ?
                                               AND $_roomIdColumn = ?;
      ''',
      [id.toString(), room.id.toString()]
    );

    if (results.isNotEmpty) {
      var content = results.first[_contentColumn];
      return Event.fromJson(room, json.decode(content));
    } else {
      return null;
    }
  }

  @override
  Stream<Event> getEvents(Room room, {int count, RoomEvent from}) async* {
    var query =
    '''
      SELECT $_eventTable.$_contentColumn,
             $_userTable.$_idColumn,
             $_userTable.$_displayNameColumn 
      FROM $_eventTable
        JOIN $_userTable
          ON $_userTable.$_idColumn=$_eventTable.$_senderIdColumn
      WHERE $_eventTable.$_roomIdColumn = ?
    ''';

    var values = <dynamic>[room.id.toString()];

    if (from != null) {
      query += ' AND $_eventTable.$_timeColumn < ?';
      values.add(from.time.millisecondsSinceEpoch);
    }

    query += ' ORDER BY $_eventTable.$_timeColumn DESC';

    if (count != null) {
      query += ' LIMIT ?';
      values.add(count);
    }

    var results = await _db.rawQuery(query, values);

    if (results.isNotEmpty) {
      for (var result in results) {
        var content = result[_contentColumn];
        var user = User(
          id: UserId(result[_idColumn]),
          name: result[_displayNameColumn]
        );

        yield await Event.fromJson(room, json.decode(content), user: user);
      }
    }
  }

  @override
  Future<void> addEvents(Room room, Iterable<Event> events) async {
    await _db.transaction((txn) {
      for (Event event in events) {
        var senderId;
        var time;

        if (event is RoomEvent) {
          senderId = event.sender.id.toString();
          time = event.time.millisecondsSinceEpoch;
        }

        txn.execute(
          '''
          INSERT OR IGNORE INTO $_eventTable($_idColumn,
                                             $_roomIdColumn,
                                             $_timeColumn,
                                             $_senderIdColumn,
                                             $_contentColumn)
            values (?, ?, ?, ?, ?);
          ''',
          [
            event.id.toString(),
            room.id.toString(),
            time,
            senderId,
            json.encode(event.toJson())]
        );
      }
    });
  }

  Future<void> _setUser(Transaction txn, User user) async {
    txn.execute(
      '''
      UPDATE $_userTable SET $_displayNameColumn = ?
                         WHERE $_idColumn = ?;
      ''',
      [user.name, user.id.toString()]
    );

    txn.execute(
      '''
      INSERT OR IGNORE INTO $_userTable($_idColumn,
                                        $_displayNameColumn)
        values (?, ?);
      ''',
      [user.id.toString(), user.name]
    );
  }

  @override
  Future<User> getMember(RoomId roomId, UserId id) async {
    var results = await _db.rawQuery(
      '''
      SELECT $_displayNameColumn FROM $_userTable
      INNER JOIN $_memberTable
      ON $_userTable.$_idColumn = $_memberTable.$_userIdColumn
      WHERE $_memberTable.$_roomIdColumn = ?
       AND $_memberTable.$_userIdColumn = ?
      ''',
      [roomId.toString(), id.toString()]
    );

    if (results.isNotEmpty) {
      var displayName = results.first[_displayNameColumn];
      return User(id: id, name: displayName);
    } else {
      return null;
    }
  }

  @override
  Stream<User> getMembers(RoomId roomId) async* {
    var results = await _db.rawQuery(
      '''
      SELECT $_userTable.$_idColumn, $_displayNameColumn FROM $_userTable
      INNER JOIN $_memberTable
      ON $_userTable.$_idColumn = $_memberTable.$_userIdColumn
      WHERE $_memberTable.$_roomIdColumn = ?
      ''',
      [roomId.toString()]
    );

    if (results.isNotEmpty) {
      for (var result in results) {
        var id = UserId(result[_idColumn]);
        var displayName = result[_displayNameColumn];

        yield User(id: id, name: displayName);
      }
    }
  }

  @override
  Future<void> addMembers(RoomId roomId, Iterable<User> members) async {
    await _db.transaction((txn) {
      for (User member in members) {
        _setUser(txn, member);

        txn.execute(
          '''
          INSERT OR IGNORE INTO $_memberTable($_userIdColumn, 
                                              $_roomIdColumn)
            values (?, ?);
          ''',
          [member.id.toString(), roomId.toString()]
        );
      }
    });
  }
}
